use adw::prelude::*;
use config::Config;
use copypasta::{ClipboardContext, ClipboardProvider};
use rbw::api::Client;
use rbw::db::{Db, Entry, EntryData, Uri};
use rbw::error::Error;
use rbw::identity::Identity;
use rbw::locked::Keys;
use relm4::component::{AsyncComponent, AsyncComponentParts, AsyncComponentSender};
use relm4::factory::{AsyncFactoryComponent, AsyncFactoryVecDeque};
use relm4::gtk::{glib, Box};
use relm4::prelude::{AsyncComponentController, DynamicIndex, SimpleAsyncComponent};
use relm4::{adw, gtk, AsyncFactorySender};
use relm4::{RelmApp, RelmWidgetExt};
use relm4_icons::icon_names;
use serde::Deserialize;
use std::clone::Clone;
use std::fmt::Debug;
use std::path::PathBuf;
use std::time::Duration;

static DEVICE_ID: &str = "bitritter";

// TODO check for dead enum members
#[derive(Debug)]
enum AppMsg {
    Startup,
    // AddVaultItem,
    SetupAccount,
    UnlockActiveAccount(String),
    LockActiveAccount,
    Sync,
    Login(String),
    // AddAccount(Account),
    SetVaultItemFilter(VaultStackState),
    ShowDetailView(String),
    // ClipboardCopy(String),
    CopyPassword(String),
}

// TODO check for dead enum members
enum AppState {
    Init,
    // AccountSet,
    DetailPage,
}

#[derive(Deserialize)]
struct Credentials {
    base_url: String,
    identity_url: String,
    email: String,
    password: String,
}

#[derive(Debug)]
enum VaultStackState {
    All,
    Login,
    Card,
    Identity,
    Note,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone)]
enum VaultItemType {
    Login,
    Card,
    Identity,
    Note,
}

struct AppModel {
    state: AppState,
    vault_stack_state: VaultStackState,
    accounts: Vec<Account>,
    active_account: Option<Account>,
    active_entry: Option<Entry>,
    // entries: HashMap<Account, Vec<Entry>>,
    items: AsyncFactoryVecDeque<ListItem>,
    detail_box: Option<Box>,
}

#[derive(PartialEq, Eq, Clone, Debug)]
enum AccountState {
    Locked,
    Unlocked,
}

fn decrypt(mut entry: Entry, unlocked_keys: &Keys) -> Result<Entry, Error> {
    // TODO implement decryption for each field
    entry.name = decrypt_string(&entry.name, unlocked_keys);
    entry.data = match entry.data {
        EntryData::Login {
            username,
            password,
            totp,
            uris,
        } => EntryData::Login {
            username: Some(decrypt_string(
                username.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            password: Some(decrypt_string(
                password.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            totp: Some(decrypt_string(
                totp.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            uris,
        },
        EntryData::Card {
            cardholder_name,
            number,
            brand,
            exp_month,
            exp_year,
            code,
        } => EntryData::Card {
            cardholder_name: Some(decrypt_string(
                cardholder_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            number: Some(decrypt_string(
                number.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            brand: Some(decrypt_string(
                brand.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            exp_month: Some(decrypt_string(
                exp_month.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            exp_year: Some(decrypt_string(
                exp_year.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            code: Some(decrypt_string(
                code.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
        },
        EntryData::Identity {
            title,
            first_name,
            middle_name,
            last_name,
            address1,
            address2,
            address3,
            city,
            state,
            postal_code,
            country,
            phone,
            email,
            ssn,
            license_number,
            passport_number,
            username,
        } => EntryData::Identity {
            title: Some(decrypt_string(
                title.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            first_name: Some(decrypt_string(
                first_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            middle_name: Some(decrypt_string(
                middle_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            last_name: Some(decrypt_string(
                last_name.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            address1: Some(decrypt_string(
                address1.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            address2: Some(decrypt_string(
                address2.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            address3: Some(decrypt_string(
                address3.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            city: Some(decrypt_string(
                city.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            state: Some(decrypt_string(
                state.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            postal_code: Some(decrypt_string(
                postal_code.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            country: Some(decrypt_string(
                country.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            phone: Some(decrypt_string(
                phone.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            email: Some(decrypt_string(
                email.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            ssn: Some(decrypt_string(
                ssn.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            license_number: Some(decrypt_string(
                license_number.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            passport_number: Some(decrypt_string(
                passport_number.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
            username: Some(decrypt_string(
                username.unwrap_or_else(String::new).as_str(),
                unlocked_keys,
            )),
        },
        EntryData::SecureNote => EntryData::SecureNote,
    };
    Ok(entry)
}

struct Account {
    base_url: String,
    identity_url: String,
    client_cert_path: Option<PathBuf>,
    email: String,
    state: AccountState,
    entries: Option<Vec<Entry>>,
    unlocked_keys: Option<Keys>,
    db: Db,
}

impl Debug for Account {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Account")
            .field("base_url", &self.base_url)
            .field("identity_url", &self.identity_url)
            .field("email", &self.email)
            .finish_non_exhaustive()
    }
}

impl Clone for Account {
    fn clone(&self) -> Self {
        Self {
            base_url: self.base_url.clone(),
            identity_url: self.identity_url.clone(),
            client_cert_path: self.client_cert_path.clone(),
            email: self.email.clone(),
            state: AccountState::Locked,
            entries: None,
            unlocked_keys: None,
            db: Db::new(),
        }
    }
}

impl Account {
    #[must_use]
    pub fn new(
        base_url: String,
        identity_url: String,
        client_cert_path: Option<PathBuf>,
        email: String,
    ) -> Account {
        Account {
            base_url,
            identity_url,
            client_cert_path,
            email,
            // keys: None,
            state: AccountState::Locked,
            entries: None,
            unlocked_keys: None,
            db: Db::new(),
        }
    }

    pub fn get_client(&self) -> Client {
        Client::new(
            self.base_url.as_str(),
            self.identity_url.as_str(),
            self.client_cert_path.as_deref(),
        )
    }

    pub async fn pre_login(&mut self) {
        let client = self.get_client();
        if let Ok((kdf_type, iterations, memory, parallelism)) =
            client.prelogin(self.email.as_str()).await
        {
            self.db.kdf = Some(kdf_type);
            self.db.iterations = Some(iterations);
            self.db.memory = memory;
            self.db.parallelism = parallelism;
        }
    }

    pub async fn login(&mut self, password: &str) -> Result<(), Error> {
        let client = self.get_client();

        let mut password_vec = rbw::locked::Vec::new();
        password_vec.extend(password.bytes());
        let password = rbw::locked::Password::new(password_vec.clone());

        let identity = Identity::new(
            self.email.as_str(),
            &password,
            self.db.kdf.expect("KDF should be set"),
            self.db.iterations.expect("iterations should be set"),
            self.db.memory,
            self.db.parallelism,
        )?;
        let (access_token, refresh_token, _key3) = client
            .login(
                self.email.as_str(),
                DEVICE_ID,
                &identity.master_password_hash,
                None,
                None,
            )
            .await?;

        self.db.access_token = Some(access_token);
        self.db.refresh_token = Some(refresh_token);
        Ok(())
    }

    pub fn unlock(&mut self, password: &str) -> Result<(), Error> {
        let mut password_vec = rbw::locked::Vec::new();
        password_vec.extend(password.bytes());

        let password = rbw::locked::Password::new(password_vec.clone());

        // TODO check if hash_map is the org_keys
        let (unlocked_keys, _unlocked_org_keys) = rbw::actions::unlock(
            &self.email,
            &password,
            self.db.kdf.expect("KDF should be set"),
            self.db.iterations.expect("iterations not set"),
            self.db.memory,
            self.db.parallelism,
            self.db
                .protected_key
                .as_ref()
                .expect("Protected key not set")
                .as_str(),
            self.db
                .protected_private_key
                .as_ref()
                .expect("Protected private key not set")
                .as_str(),
            &self.db.protected_org_keys,
        )?;
        // TODO unlock and decrypt entries with the unlock key
        self.decrypt(&unlocked_keys);
        self.unlocked_keys = Some(unlocked_keys);
        self.state = AccountState::Unlocked;
        Ok(())
    }

    pub fn lock(&mut self) {
        self.unlocked_keys = None;
        self.entries = None;
        self.state = AccountState::Locked;
    }

    pub async fn sync(&mut self) {
        let client = self.get_client();
        let (protected_key, protected_private_key, _protected_org_keys, entries) = client
            .sync(
                self.db
                    .access_token
                    .as_ref()
                    .expect("access token needed for sync")
                    .as_str(),
            )
            .await
            .unwrap();

        self.db.protected_key = Some(protected_key);
        self.db.protected_private_key = Some(protected_private_key);
        // self.db.protected_org_keys = Some(protected_org_keys);
        self.db.entries = entries;
    }

    // TODO replace by db.entries?
    pub fn decrypt(&mut self, unlocked_keys: &Keys) {
        self.entries = Some(
            self.db
                .entries
                .clone()
                .iter()
                .map(|entry| decrypt(entry.clone(), unlocked_keys).expect("TODO"))
                .collect(),
        );
        // TODO part splitter
        // if let Some(entries) = &self.entries {
        //     let decrypted_entries = entries.iter().map(
        //         |entry| decrypt(entry.clone(), &unlocked_keys).expect("TODO"), // TODO
        //     );
        //     self.entries = Some(decrypted_entries.collect());
        // }
    }

    pub async fn save_to_db(&self) -> Result<(), Error> {
        self.db.save_async(&self.base_url, &self.email).await
    }

    pub async fn load_from_db(&mut self) {
        if let Ok(db) = Db::load_async(&self.base_url, &self.email).await {
            self.db = db;
        }
    }
}

fn decrypt_string(string_to_decrypt: &str, unlocked_keys: &Keys) -> String {
    if !string_to_decrypt.is_empty() {
        let x: &[_] = &[
            '\t', '\n', '\r', '\u{1}', '\u{2}', '\u{3}', '\u{4}', '\u{5}', '\u{6}', '\u{7}',
            '\u{8}', '\u{10}', '\u{b}', '\u{c}', '\u{e}', '\u{f}',
        ];
        let cipher_id: rbw::cipherstring::CipherString =
            rbw::cipherstring::CipherString::new(string_to_decrypt).unwrap();
        let decrypted_id = cipher_id
            .decrypt_locked_symmetric(unlocked_keys)
            .inspect_err(|e| eprintln!("failed decrypting id: {e}"));
        std::str::from_utf8(decrypted_id.unwrap().data())
            .unwrap()
            .trim_matches(x)
            .to_string()
    } else {
        String::new()
    }
}

// TODO
#[derive(Clone)]
struct ListItem {
    id: String,
    primary: String,
    secondary: String,
    item_type: VaultItemType,
}

#[derive(Debug)]
enum ListOutput {
    // Select(DynamicIndex),
    Select(String),
    CopyPassword(String),
}

#[relm4::factory(async)]
impl AsyncFactoryComponent for ListItem {
    type Init = ListItem;
    type Input = ();
    type Output = ListOutput;
    type CommandOutput = ();
    type ParentWidget = gtk::ListBox;

    view! {
        root = gtk::ListBox {
            adw::ActionRow {
                set_title: self.primary.as_str(),
                set_subtitle: self.secondary.as_str(),
                set_activatable: true,
                set_selectable: false,
                add_prefix = &gtk::Image {
                    set_icon_name: Some( match self.item_type {
                        VaultItemType::Login => icon_names::PERMISSIONS_PASSWORD,
                        VaultItemType::Card => icon_names::VCARD,
                        VaultItemType::Identity => icon_names::PERSON,
                        VaultItemType::Note => icon_names::NOTE
                    })
                },
                connect_activated[sender, model = self.clone()] => move |_| {
                     let _ = sender.output(ListOutput::Select(model.id.clone()));
                },
                set_hexpand: true,
                add_suffix = match self.item_type {
                    VaultItemType::Login => {
                        &gtk::Button {
                            set_icon_name: icon_names::PASSWORD_ENTRY,
                            set_tooltip: "Copy Password",
                            connect_clicked[sender, model = self.clone()] => move |_| {
                                let _ = sender.output(ListOutput::CopyPassword(model.id.clone()));
                            }
                        }
                    },
                    VaultItemType::Card => {
                        &gtk::Button {
                          set_label: "Card"
                        }
                    },
                    VaultItemType::Identity => {
                        &gtk::Button {
                          set_label: "Identity"
                        }
                    },
                    VaultItemType::Note => {
                        &gtk::Button {
                          set_label: "Note"
                        }
                    }
                }
            }
        }
    }

    async fn init_model(
        item: Self::Init,
        _index: &DynamicIndex,
        _sender: AsyncFactorySender<Self>,
    ) -> Self {
        tokio::time::sleep(Duration::from_secs(1)).await;
        Self {
            id: item.id,
            primary: item.primary,
            secondary: item.secondary,
            item_type: item.item_type,
        }
    }

    async fn update(&mut self, msg: Self::Input, _sender: AsyncFactorySender<Self>) {
        tokio::time::sleep(Duration::from_secs(1)).await;
    }
}

struct DetailCard {
    id: String,
    name: String,
    cardholder_name: Option<String>,
    number: Option<String>,
    brand: Option<String>,
    exp_month: Option<String>,
    exp_year: Option<String>,
    code: Option<String>,
}

#[relm4::component(async)]
impl SimpleAsyncComponent for DetailCard {
    type Init = Entry;
    type Input = ();
    type Output = ();

    view! {
        #[root]
        root_box = &gtk::Box {
            set_orientation: gtk::Orientation::Vertical,
            set_hexpand:true,

            gtk::Label {
                set_text: model.name.as_ref()
            },
            adw::ActionRow {
                set_title: "Cardholder Name",
                set_subtitle: model.cardholder_name.as_ref().unwrap_or(&String::new())
            },
            adw::PasswordEntryRow {
                set_title: "Card Number",
                set_text: model.number.as_ref().unwrap_or(&String::new())
            },
            adw::PasswordEntryRow {
                set_title: "Security Code",
                set_text: model.code.as_ref().unwrap_or(&String::new())
            },
            // TODO add icons?
            adw::ActionRow {
                set_title: "Brand",
                set_subtitle: model.brand.as_ref().unwrap_or(&String::new())
            },
            // TODO format it?
            adw::ActionRow {
                set_title: "Month of Expiration",
                set_subtitle: model.exp_month.as_ref().unwrap_or(&String::new())
            },
            adw::ActionRow {
                set_title: "Year of Expiration",
                set_subtitle: model.exp_year.as_ref().unwrap_or(&String::new())
            },
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login { .. } => todo!(),
            EntryData::Card {
                cardholder_name,
                number,
                brand,
                exp_month,
                exp_year,
                code,
            } => DetailCard {
                id: entry_data.id,
                name: entry_data.name,
                cardholder_name,
                number,
                brand,
                exp_month,
                exp_year,
                code,
            },
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => todo!(),
        };
        let widgets = view_output!();

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, message: Self::Input, sender: AsyncComponentSender<Self>) {}
}

struct DetailLogin {
    id: String,
    name: String,
    username: Option<String>,
    password: Option<String>,
    totp: Option<String>,
    uris: Vec<Uri>,
}

#[relm4::component(async)]
impl SimpleAsyncComponent for DetailLogin {
    type Init = Entry;
    type Input = ();
    type Output = ();

    view! {
        #[root]
        root_box = &gtk::Box {
            set_orientation: gtk::Orientation::Vertical,
            set_hexpand:true,

            gtk::Label {
                set_text: model.name.as_ref()
            },
            adw::ActionRow {
                set_title: "Username",
                set_subtitle: model.username.as_ref().unwrap_or(&String::new())
            },
            adw::PasswordEntryRow {
                set_title: "Password",
                set_text: model.password.as_ref().unwrap_or(&String::new())
            },
            adw::ActionRow {
                set_title: "TOTP",
                set_visible: model.totp.as_ref().is_some(),
                set_subtitle: model.totp.as_ref().unwrap_or(&String::new())
            },
            #[name = "uris_anchor"]
            gtk::Box {}
        }
    }

    async fn init(
        entry_data: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        let model = match entry_data.data {
            EntryData::Login {
                username,
                password,
                totp,
                uris,
            } => DetailLogin {
                id: entry_data.id,
                name: entry_data.name,
                username,
                password,
                totp,
                uris,
            },
            EntryData::Card { .. } => todo!(),
            EntryData::Identity { .. } => todo!(),
            EntryData::SecureNote => todo!(),
        };

        let widgets = view_output!();

        for uri in model.uris.iter() {
            widgets
                .uris_anchor
                .append(&adw::ActionRow::builder().subtitle(uri.uri.clone()).build());
        }

        AsyncComponentParts { model, widgets }
    }

    async fn update(&mut self, message: Self::Input, sender: AsyncComponentSender<Self>) {}
}

#[relm4::component(async)]
impl AsyncComponent for AppModel {
    type Init = AppState;
    type Input = AppMsg;
    type Output = ();
    type CommandOutput = ();

    view! {
        #[root]
        main_window = &adw::Window {
            set_default_width: 300,
            set_default_height: 500,

            #[name = "main_stack"]
            gtk::Stack {
                #[transition = "SlideRight"]
                add_child = if model.active_account.is_none() {
                    #[name = "account_chooser"]
                    gtk::Box {
                        set_orientation: gtk::Orientation::Vertical,

                        adw::HeaderBar {
                            #[wrap(Some)]
                            set_title_widget = &gtk::Label {
                                set_text: "Set up an account"
                            }
                        },
                        gtk::Button {
                            set_label: "+",
                            connect_clicked[sender] => move |_| {
                                // TODO add account creation here
                                sender.input(AppMsg::SetupAccount);
                            }
                        },
                    }
                } else {
                    #[name = "leaflet"]
                    adw::Leaflet {
                        set_can_navigate_back: true,
                        set_vexpand: true,
                        set_visible_child_name: if model.active_entry.is_some() { "detail" } else { "list" },

                        #[name = "list"]
                        gtk::Box {
                            set_orientation: gtk::Orientation::Vertical,

                            #[name = "list_headerbar"]
                            adw::HeaderBar {
                                #[wrap(Some)]
                                set_title_widget = &gtk::Box {
                                    set_orientation: gtk::Orientation::Horizontal,
                                    gtk::Button {
                                        set_icon_name: "open-menu",
                                        set_halign: gtk::Align::Start
                                    },
                                    gtk::Button {
                                        set_icon_name: "system-search"
                                    },
                                    gtk::Button {
                                        set_label: "Login",
                                        connect_clicked[sender] => move |_| {
                                            // TODO remove password and read from prompt
                                            sender.input(AppMsg::Login(load_password()));
                                        }
                                    },
                                    gtk::Button {
                                        set_label: "Unlock",
                                        connect_clicked[sender] => move |_| {
                                            // TODO remove password and read from prompt
                                            sender.input(AppMsg::UnlockActiveAccount(load_password()));
                                        }
                                    },
                                    gtk::Button {
                                        set_label: "Sync",
                                        connect_clicked[sender] => move |_| {
                                            sender.input(AppMsg::Sync);
                                        }
                                    },
                                    gtk::Button {
                                        set_label: "Load",
                                        connect_clicked[sender] => move |_| {
                                            sender.input(AppMsg::Startup);
                                        }
                                    }
                                }
                            },
                            append: password_input = &gtk::Box {
                                set_orientation: gtk::Orientation::Vertical,
                                #[watch]
                                set_visible:  model.active_account.as_ref().is_some_and(|acc| acc.state == AccountState::Locked),
                                gtk::Label {
                                    set_text: "Enter password to unlock Vault"
                                },
                                #[name = "password_entry"]
                                adw::PasswordEntryRow {
                                    set_title: "Password"
                                },
                                gtk::Button {
                                    set_label: "Unlock Vault",
                                    connect_clicked[sender, password_entry] => move |_| {
                                        sender.input(AppMsg::UnlockActiveAccount(password_entry.text().to_string()))
                                    }
                                }
                            },
                            gtk::ScrolledWindow {
                                set_vexpand: true,
                                #[local_ref]
                                // This is just an anchor
                                list_box -> gtk::ListBox {}
                            },
                            gtk::Box {
                                set_orientation: gtk::Orientation::Horizontal,
                                set_hexpand: true,

                                gtk::Button {
                                    set_label: "All",
                                    set_hexpand: true,
                                    set_margin_horizontal: 3,
                                    connect_clicked[sender] => move |_| {
                                        sender.input(AppMsg::SetVaultItemFilter(VaultStackState::All));
                                    }
                                },
                                gtk::Button {
                                    set_label: "Login",
                                    set_hexpand: true,
                                    set_margin_horizontal: 3,
                                    connect_clicked[sender] => move |_| {
                                        sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Login));
                                    }
                                },
                                gtk::Button {
                                    set_label: "Card",
                                    set_hexpand: true,
                                    set_margin_horizontal: 3,
                                    connect_clicked[sender] => move |_| {
                                        sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Card));
                                    }
                                },
                                gtk::Button {
                                    set_label: "Identity",
                                    set_hexpand: true,
                                    set_margin_horizontal: 3,
                                    connect_clicked[sender] => move |_| {
                                        sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Identity));
                                    }
                                },
                                gtk::Button {
                                    set_label: "Note",
                                    set_hexpand: true,
                                    set_margin_horizontal: 3,
                                    connect_clicked[sender] => move |_| {
                                        sender.input(AppMsg::SetVaultItemFilter(VaultStackState::Note));
                                    }
                                },
                            },
                        },
                        #[name = "detail"]
                        gtk::Box {
                            set_orientation: gtk::Orientation::Vertical,
                            set_hexpand: true,
                            #[name = "detail_headerbar"]
                            adw::HeaderBar {
                                #[name = "back_button"]
                                pack_start = &gtk::Button {
                                    set_icon_name: "go-previous-symbolic",
                                    connect_clicked[sender] => move |_| {
                                        sender.input(AppMsg::ShowDetailView(String::new()));
                                    },
                                }
                            },
                            #[name = "detail_anchor"]
                            gtk::Box {
                                set_vexpand: true,
                                set_hexpand: true,
                            },
                        }
                    }
                }
            }
        }
    }

    // menu! {}

    async fn init(
        _state: Self::Init,
        root: Self::Root,
        sender: AsyncComponentSender<Self>,
    ) -> AsyncComponentParts<Self> {
        tokio::time::sleep(Duration::from_millis(100)).await;

        let items = AsyncFactoryVecDeque::builder().launch_default().forward(
            sender.input_sender(),
            |output| match output {
                ListOutput::Select(id) => AppMsg::ShowDetailView(id),
                ListOutput::CopyPassword(id) => {
                    let mut ctx = ClipboardContext::new().unwrap();
                    ctx.set_contents(id.to_owned()).unwrap();
                    AppMsg::CopyPassword(id)
                }
            },
        );

        let mut loaded_active_account = load_active_account().await;
        if let Some(ref mut account) = loaded_active_account {
            account.load_from_db().await;
        }

        let model = AppModel {
            state: AppState::Init,
            vault_stack_state: VaultStackState::All,
            accounts: Vec::new(),
            active_account: loaded_active_account,
            active_entry: None,
            // entries: HashMap::new(),
            items,
            detail_box: None,
        };

        let list_box = model.items.widget();
        // list_box.set_filter_func(|item| match item.downcast::<ListItem>() {
        //     Ok(list_item) => list_item.item_type == VaultItemType::Login
        //     Err(_) => todo!(),
        // });

        let widgets = view_output!();

        // Bindings for the Leaflet interaction
        widgets
            .leaflet
            .bind_property("folded", &widgets.list_headerbar, "show-end-title-buttons")
            .flags(glib::BindingFlags::SYNC_CREATE)
            .build();

        widgets
            .leaflet
            .bind_property(
                "folded",
                &widgets.detail_headerbar,
                "show-start-title-buttons",
            )
            .flags(glib::BindingFlags::SYNC_CREATE)
            .build();

        widgets
            .leaflet
            .bind_property("folded", &widgets.back_button, "visible")
            .flags(glib::BindingFlags::SYNC_CREATE)
            .build();

        AsyncComponentParts { model, widgets }
    }

    async fn update_with_view(
        &mut self,
        widgets: &mut AppModelWidgets,
        message: Self::Input,
        sender: AsyncComponentSender<Self>,
        _root: &Self::Root,
    ) {
        tokio::time::sleep(Duration::from_millis(100)).await;

        match message {
            AppMsg::Startup => {
                if let Some(ref mut account) = self.active_account {
                    account.load_from_db().await;
                }
            }
            AppMsg::Login(password) => {
                if let Some(ref mut account) = self.active_account {
                    account.pre_login().await;
                    // if let Ok(_) = account.login(password.as_str()).await {
                    if (account.login(password.as_str()).await).is_ok() {
                        let _save_result = account.save_to_db().await;
                    }
                    println!("Logged in");
                }
            }
            AppMsg::Sync => {
                if let Some(ref mut account) = self.active_account {
                    account.sync().await;
                    let _save_result = account.save_to_db().await;
                    println!("Sync complete");
                }
            }
            AppMsg::UnlockActiveAccount(password) => {
                if let Some(ref mut account) = self.active_account {
                    match account.unlock(password.as_str()) {
                        // TODO open error dialog here
                        Err(_) => todo!(),
                        Ok(_) => {
                            for entry in account.entries.to_owned().unwrap() {
                                // println!("Unlocking: {}", entry.id);
                                self.items.guard().insert(
                                    0,
                                    ListItem {
                                        id: entry.id,
                                        item_type: match entry.data {
                                            EntryData::Login { .. } => VaultItemType::Login,
                                            EntryData::Card { .. } => VaultItemType::Card,
                                            EntryData::Identity { .. } => VaultItemType::Identity,
                                            EntryData::SecureNote => VaultItemType::Note,
                                        },
                                        primary: if !entry.name.is_empty() {
                                            entry.name
                                        } else {
                                            "empty".to_string()
                                        },
                                        secondary: match entry.data {
                                            EntryData::Login { username, .. } => {
                                                if username.clone().is_some_and(|n| !n.is_empty()) {
                                                    username.unwrap()
                                                } else {
                                                    String::new()
                                                }
                                            }
                                            EntryData::Card {
                                                cardholder_name,
                                                .. /*
                                                number,
                                                brand,
                                                exp_month,
                                                exp_year,
                                                code, */
                                            } => {
                                                if cardholder_name.clone().is_some_and(|n| !n.is_empty()) {
                                                    cardholder_name.unwrap()
                                                } else {
                                                    String::new()
                                                }
                                            }
                                            // TODO correct subtitle pls
                                            EntryData::Identity {
                                                .. /*
                                                title,
                                                first_name,
                                                middle_name,
                                                last_name,
                                                address1,
                                                address2,
                                                address3,
                                                city,
                                                state,
                                                postal_code,
                                                country,
                                                phone,
                                                email,
                                                ssn,
                                                license_number,
                                                passport_number,
                                                username, */
                                            } => String::new(),
                                            EntryData::SecureNote => String::new(),
                                        },
                                    },
                                );
                            }
                            println!("Unlock complete");
                            self.update_view(widgets, sender);
                        }
                    }
                }
            }
            AppMsg::LockActiveAccount => {
                if let Some(ref mut acc) = self.active_account {
                    acc.lock();
                }
            }

            // TODO check for dead enum members
            // AppMsg::AddAccount(account) => {}

            // AppMsg::AddVaultItem => {}
            AppMsg::SetupAccount => {
                println!("Setup Account");
            }
            AppMsg::SetVaultItemFilter(filter) => match filter {
                VaultStackState::Login => {
                    self.vault_stack_state = VaultStackState::Login;
                }
                VaultStackState::Card => {
                    self.vault_stack_state = VaultStackState::Card;
                }
                VaultStackState::Identity => {
                    self.vault_stack_state = VaultStackState::Identity;
                }
                VaultStackState::Note => {
                    self.vault_stack_state = VaultStackState::Note;
                }
                _ => {}
            },
            AppMsg::ShowDetailView(id) => {
                if id.is_empty() {
                    widgets.leaflet.navigate(adw::NavigationDirection::Back);
                } else {
                    widgets.leaflet.navigate(adw::NavigationDirection::Forward);
                    self.state = AppState::DetailPage;

                    // TODO prettify those unwraps
                    let new_entry = self
                        .active_account
                        .as_ref()
                        .unwrap()
                        .entries
                        .as_ref()
                        .unwrap()
                        .clone()
                        .into_iter()
                        .find(|entry| entry.id == id)
                        .unwrap()
                        .clone();
                    self.active_entry = Some(new_entry.clone());

                    if let Some(ref detail_widget) = self.detail_box {
                        widgets.detail_anchor.remove(detail_widget);
                    };
                    match new_entry.data {
                        EntryData::Login { .. } => {
                            let detail_login = DetailLogin::builder().launch(new_entry).detach();
                            widgets.detail_anchor.append(detail_login.widget());
                            self.detail_box = Some(detail_login.widget().to_owned());
                        }
                        EntryData::Card { .. } => {
                            let detail_card = DetailCard::builder().launch(new_entry).detach();
                            widgets.detail_anchor.append(detail_card.widget());
                            self.detail_box = Some(detail_card.widget().to_owned());
                        }
                        EntryData::Identity { .. } => todo!(),
                        EntryData::SecureNote => todo!(),
                    };
                    self.update_view(widgets, sender);
                }
            }
            AppMsg::CopyPassword(id) => {
                let entry = self
                    .active_account
                    .as_ref()
                    .unwrap()
                    .entries
                    .clone()
                    .unwrap()
                    .iter()
                    .find(|entry| entry.id == id)
                    .unwrap()
                    .to_owned();
                let pass = match entry.data.clone() {
                    EntryData::Login {
                        password,
                        ..
                        /*
                        username,
                        totp,
                        uris, */
                    } => password.unwrap(),
                    // decrypt_string(
                    //     password.unwrap().as_str(),
                    //     &self
                    //         .active_account
                    //         .as_ref()
                    //         .unwrap()
                    //         .unlocked_keys
                    //         .clone()
                    //         .unwrap(),
                    // ),
                    EntryData::Card { .. } => String::new(),
                    EntryData::Identity { .. } => String::new(),
                    EntryData::SecureNote => String::new(),
                };
                let mut ctx = ClipboardContext::new().unwrap();
                ctx.set_contents(pass.to_owned()).unwrap();
                let _clip = ctx.get_contents().unwrap();
            }
        }
    }
}

fn load_password() -> String {
    let credentials = Config::builder()
        .add_source(config::File::with_name("credentials.toml"))
        .build()
        .unwrap();

    let cred = credentials.try_deserialize::<Credentials>().unwrap();

    cred.password
}

async fn load_active_account() -> Option<Account> {
    // Right now this is just a mock method to get users started quickly.
    // Later, all account management should happen in the UI and be stored
    // in a DB/own file structure
    let credentials = Config::builder()
        .add_source(config::File::with_name("credentials.toml"))
        .build()
        .unwrap();

    match credentials.try_deserialize::<Credentials>() {
        Ok(cred) => Some(Account::new(
            cred.base_url,
            cred.identity_url,
            None,
            cred.email,
        )),
        Err(_) => None,
    }
}

fn main() {
    gtk::init().unwrap();
    let app = RelmApp::new("dev.bitritter");
    relm4_icons::initialize_icons();
    app.run_async::<AppModel>(AppState::Init);
}
